from time import time
from typing import Generator, Optional

from httpx import Auth as HttpxAuth
from httpx import Request, Response, codes

from firebolt.client.constants import _REQUEST_ERRORS, DEFAULT_API_URL
from firebolt.common.exception import AuthenticationError
from firebolt.common.token_storage import TokenSecureStorage
from firebolt.common.urls import AUTH_URL
from firebolt.common.util import fix_url_schema


class Auth(HttpxAuth):

    """
    Authentication class for Firebolt database. Gets authentication token using
    provided credentials and updates it when it expires
    """

    __slots__ = (
        "username",
        "password",
        "api_url",
        "_token",
        "_expires",
    )

    requires_response_body = True

    @staticmethod
    def from_token(token: str) -> "Auth":
        a = Auth("", "")
        a._token = token
        return a

    def __init__(
        self,
        username: str,
        password: str,
        api_endpoint: str = DEFAULT_API_URL,
    ):
        self.username = username
        self.password = password
        self._token_storage = TokenSecureStorage(username=username, password=password)

        # Add schema to url if it's missing
        self._api_endpoint = fix_url_schema(api_endpoint)
        self._token: Optional[str] = self._token_storage.get_cached_token()
        self._expires: Optional[int] = None

    def copy(self) -> "Auth":
        return Auth(self.username, self.password, self._api_endpoint)

    @property
    def token(self) -> Optional[str]:
        return self._token

    @property
    def expired(self) -> Optional[int]:
        return self._expires is not None and self._expires <= int(time())

    def get_new_token_generator(self) -> Generator[Request, Response, None]:
        """Get new token using username and password"""
        try:
            response = yield Request(
                "POST",
                AUTH_URL.format(api_endpoint=self._api_endpoint),
                headers={
                    "Content-Type": "application/json;charset=UTF-8",
                    "User-Agent": "firebolt-sdk",
                },
                json={"username": self.username, "password": self.password},
            )
            response.raise_for_status()

            parsed = response.json()
            self._check_response_error(parsed)

            self._token = parsed["access_token"]
            self._expires = int(time()) + int(parsed["expires_in"])

            self._token_storage.cache_token(parsed["access_token"], self._expires)

        except _REQUEST_ERRORS as e:
            raise AuthenticationError(repr(e), self._api_endpoint)

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        """Add authorization token to request headers.
        Overrides ``httpx.Auth.auth_flow``"""

        if not self.token or self.expired:
            yield from self.get_new_token_generator()

        request.headers["Authorization"] = f"Bearer {self.token}"

        response = yield request

        if response.status_code == codes.UNAUTHORIZED:
            yield from self.get_new_token_generator()
            request.headers["Authorization"] = f"Bearer {self.token}"
            yield request

    def _check_response_error(self, response: dict) -> None:
        if "error" in response:
            raise AuthenticationError(
                response.get("message", "unknown server error"),
                self._api_endpoint,
            )



from asyncio import (
    AbstractEventLoop,
    get_event_loop,
    new_event_loop,
    set_event_loop,
)
from functools import lru_cache, wraps
from threading import Thread
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Optional,
    Type,
    TypeVar,
)

T = TypeVar("T")


def cached_property(func: Callable[..., T]) -> T:
    return property(lru_cache()(func))  # type: ignore


def prune_dict(d: dict) -> dict:
    """Prune items from dictionaries where value is None"""
    return {k: v for k, v in d.items() if v is not None}


TMix = TypeVar("TMix")


def mixin_for(baseclass: Type[TMix]) -> Type[TMix]:
    """
    Useful function to make mixins with baseclass typehint
    Should be used as a mixin base class to fix typehints

    ```
    class ReadonlyMixin(mixin_for(BaseClass))):
        ...
    ```
    """

    if TYPE_CHECKING:
        return baseclass
    return object


def fix_url_schema(url: str) -> str:
    return url if url.startswith("http") else f"https://{url}"


class AsyncJobThread:
    """
    Thread runner that allows running async tasks syncronously in a separate thread.
    Caches loop to be reused in all threads
    It allows running async functions syncronously inside a running event loop.
    Since nesting loops is not allowed, we create a separate thread for a new event loop
    """

    def __init__(self) -> None:
        self.loop: Optional[AbstractEventLoop] = None
        self.result: Optional[Any] = None
        self.exception: Optional[BaseException] = None

    def _initialize_loop(self) -> None:
        if not self.loop:
            try:
                # despite the docs, this function fails if no loop is set
                self.loop = get_event_loop()
            except RuntimeError:
                self.loop = new_event_loop()
        set_event_loop(self.loop)

    def run(self, coro: Coroutine) -> None:
        try:
            self._initialize_loop()
            assert self.loop is not None
            self.result = self.loop.run_until_complete(coro)
        except BaseException as e:
            self.exception = e

    def execute(self, coro: Coroutine) -> Any:
        thread = Thread(target=self.run, args=[coro])
        thread.start()
        thread.join()
        if self.exception:
            raise self.exception
        return self.result


def async_to_sync(f: Callable, async_job_thread: AsyncJobThread = None) -> Callable:
    @wraps(f)
    def sync(*args: Any, **kwargs: Any) -> Any:
        try:
            loop = get_event_loop()
        except RuntimeError:
            loop = new_event_loop()
            set_event_loop(loop)
        # We are inside a running loop
        if loop.is_running():
            nonlocal async_job_thread
            if not async_job_thread:
                async_job_thread = AsyncJobThread()
            return async_job_thread.execute(f(*args, **kwargs))
        return loop.run_until_complete(f(*args, **kwargs))

    return sync



from pydantic import BaseSettings, Field, SecretStr, root_validator


class Settings(BaseSettings):
    # Authorization
    user: str = Field(None, env="FIREBOLT_USER")
    password: SecretStr = Field(None, env="FIREBOLT_PASSWORD")
    # Or
    access_token: str = Field(None, env="FIREBOLT_AUTH_TOKEN")

    account_name: str = Field(None, env="FIREBOLT_ACCOUNT")
    server: str = Field(..., env="FIREBOLT_SERVER")
    default_region: str = Field(..., env="FIREBOLT_DEFAULT_REGION")

    class Config:
        env_file = ".env"

    @root_validator
    def mutual_exclusive_with_creds(cls, values: dict) -> dict:
        if values["user"] or values["password"]:
            if values["access_token"]:
                raise ValueError("Provide only one of user/password or access_token")
        elif not values["access_token"]:
            raise ValueError("Provide either user/password or access_token")
        return values



class FireboltError(Exception):
    pass


class FireboltEngineError(FireboltError):
    """Base error for engine errors."""


class EngineNotRunningError(FireboltEngineError):
    pass


class NoAttachedDatabaseError(FireboltEngineError):
    def __init__(self, method_name: str):
        self.method_name = method_name

    def __str__(self) -> str:
        return (
            f"Unable to call {self.method_name}: "
            "Engine must to be attached to a database first."
        )


class AlreadyBoundError(FireboltEngineError):
    pass


class FireboltDatabaseError(FireboltError):
    pass


class AccountNotFoundError(FireboltError):
    def __init__(self, method_name: str):
        self.method_name = method_name

    def __str__(self) -> str:
        return f'Account "{self.method_name}" does not exist.'


class AttachedEngineInUseError(FireboltDatabaseError):
    def __init__(self, method_name: str):
        self.method_name = method_name

    def __str__(self) -> str:
        return (
            f"Unable to call {self.method_name}: "
            "Engine must not be in starting or stopping state."
        )


class ConnectionError(FireboltError):
    pass


class ConnectionClosedError(ConnectionError):
    pass


class CursorError(FireboltError):
    pass


class CursorClosedError(CursorError):
    def __init__(self, method_name: str):
        self.method_name = method_name

    def __str__(self) -> str:
        return f"Unable to call {self.method_name}: cursor closed."


class QueryNotRunError(CursorError):
    def __init__(self, method_name: str):
        self.method_name = method_name

    def __str__(self) -> str:
        return f"Unable to call {self.method_name}: need to run a query first."


class QueryError(CursorError):
    pass


class AuthenticationError(FireboltError):
    """
    Firebolt authentication error. Stores error cause and authentication endpoint.
    """

    def __init__(self, cause: str, api_endpoint: str):
        self.cause = cause
        self.api_endpoint = api_endpoint

    def __str__(self) -> str:
        return f"Failed to authenticate at {self.api_endpoint}: {self.cause}."


# PEP-249


class Warning(Exception):
    """
    Exception raised for important warnings
    like data truncations while inserting, etc.
    """


Error = FireboltError


class InterfaceError(Error):
    """
    Exception raised for errors that are related to the database interface
    rather than the database itself.
    """


class DatabaseError(Error):
    """Exception raised for errors that are related to the database."""


class DataError(DatabaseError):
    """
    Exception raised for errors that are due to problems with the processed data
    like division by zero, numeric value out of range, etc.
    """


class OperationalError(DatabaseError):
    """
    Exception raised for errors that are related to the database's operation
    and not necessarily under the control of the programmer, e.g. an unexpected
    disconnect occurs, the data source name is not found, a transaction could not
    be processed, a memory allocation error occurred during processing, etc.
    """


class IntegrityError(DatabaseError):
    """
    Exception raised when the relational integrity of the database is affected,
    e.g. a foreign key check fails.
    """


class InternalError(DatabaseError):
    """
    Exception raised when the database encounters an internal error,
    e.g. the cursor is not valid anymore, the transaction is out of sync, etc.
    """


class ProgrammingError(DatabaseError):
    """
    Exception raised when the database encounters an internal error,
    e.g. the cursor is not valid anymore, the transaction is out of sync, etc.
    """


class NotSupportedError(DatabaseError):
    """
    Exception raised when the database encounters an internal error,
    e.g. the cursor is not valid anymore, the transaction is out of sync, etc.
    """


class ConfigurationError(InterfaceError):
    """
    Exception raised when provided configuration is not correct
    """



from base64 import b64decode, b64encode, urlsafe_b64encode
from hashlib import sha256
from json import JSONDecodeError
from json import dump as json_dump
from json import load as json_load
from os import makedirs, path, urandom
from time import time
from typing import Optional

from appdirs import user_data_dir
from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.backends import default_backend  # type: ignore
from cryptography.hazmat.primitives import hashes  # type: ignore
from cryptography.hazmat.primitives.kdf.pbkdf2 import (
    PBKDF2HMAC,  # type: ignore
)

APPNAME = "firebolt"


def generate_salt() -> str:
    return b64encode(urandom(16)).decode("ascii")


def generate_file_name(username: str, password: str) -> str:
    username_hash = sha256(username.encode("utf-8")).hexdigest()[:32]
    password_hash = sha256(password.encode("utf-8")).hexdigest()[:32]

    return f"{username_hash}{password_hash}.json"


class TokenSecureStorage:
    def __init__(self, username: str, password: str):
        """
        Class for permanent storage of token in the filesystem in encrypted way

        :param username: username used for toke encryption
        :param password: password used for toke encryption
        """
        self._data_dir = user_data_dir(appname=APPNAME)
        makedirs(self._data_dir, exist_ok=True)

        self._token_file = path.join(
            self._data_dir, generate_file_name(username, password)
        )

        self.salt = self._get_salt()
        self.encrypter = FernetEncrypter(self.salt, username, password)

    def _get_salt(self) -> str:
        """
        Get salt from the file if exists, or generate a new one

        :return: salt
        """
        res = self._read_data_json()
        return res.get("salt", generate_salt())

    def _read_data_json(self) -> dict:
        """
        Read json token file

        :return: json object as dict
        """
        if not path.exists(self._token_file):
            return {}

        with open(self._token_file) as f:
            try:
                return json_load(f)
            except JSONDecodeError:
                return {}

    def get_cached_token(self) -> Optional[str]:
        """
        Get decrypted token using username and password
        If the token not found or token cannot be decrypted using username, password
        None will be returned

        :return: token or None
        """
        res = self._read_data_json()
        if "token" not in res:
            return None

        # Ignore expired tokens
        if "expiration" in res and res["expiration"] <= int(time()):
            return None

        return self.encrypter.decrypt(res["token"])

    def cache_token(self, token: str, expiration_ts: int) -> None:
        """

        :param token:
        :return:
        """
        token = self.encrypter.encrypt(token)

        with open(self._token_file, "w") as f:
            json_dump(
                {"token": token, "salt": self.salt, "expiration": expiration_ts}, f
            )


class FernetEncrypter:
    def __init__(self, salt: str, username: str, password: str):
        """

        :param salt:
        :param username:
        :param password:
        """

        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            salt=b64decode(salt),
            length=32,
            iterations=39000,
            backend=default_backend(),
        )
        self.fernet = Fernet(
            urlsafe_b64encode(
                kdf.derive(bytes(f"{username}{password}", encoding="utf-8"))
            )
        )

    def encrypt(self, data: str) -> str:
        return self.fernet.encrypt(bytes(data, encoding="utf-8")).decode("utf-8")

    def decrypt(self, data: str) -> Optional[str]:
        try:
            return self.fernet.decrypt(bytes(data, encoding="utf-8")).decode("utf-8")
        except InvalidToken:
            return None
